------------------------------------------
--Set up Preferences file
------------------------------------------
--slow on whole track so needs process slicer
require "process_slicer"   


--create preferences xml
local options = renoise.Document.create {
  preserve_column_pos = true
}

 --assign options-object to .preferences so renoise knows to load and update it with the tool
renoise.tool().preferences = options
------------------------------------------
------------------------------------------

--keybindings
 renoise.tool():add_keybinding {
  name = "Global:Tools:Reorder Notes By Pitch in Pattern Track",
  invoke = function() process_single_track_in_pattern() end
}

 renoise.tool():add_keybinding {
  name = "Global:Tools:Reorder Notes By Pitch in Whole Track",
  invoke = function() process_whole_track() end
}

 renoise.tool():add_keybinding {
  name = "Global:Tools:Reorder Notes By Pitch in Selected Patterns",
  invoke = function() process_whole_song() end
}


--toggle all slots selection in sequencer
 renoise.tool():add_keybinding {
  name = "Global:Tools:Toggle All Slots Selection in Sequencer",
  invoke = function() select_all_slots_in_sequencer() end
}
--menus
renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:Ledger`s Scripts:Reorder Notes By Pitch In:Pattern Track",
  invoke = function() process_single_track_in_pattern() end
}

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:Ledger`s Scripts:Reorder Notes By Pitch In:Whole Track",
  invoke = function()  process_whole_track() end
}

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:Ledger`s Scripts:Reorder Notes By Pitch In:Selected Patterns",
  invoke = function() process_whole_song() end
}

--add a preference menu entry: Preserve Column Positions
renoise.tool():add_menu_entry {
  name = "--- Main Menu:Tools:Ledger`s Scripts:Reorder Notes By Pitch In:Preserve Column Positions Enabled",
  selected = function() return options.preserve_column_pos.value end,
  invoke = function()options.preserve_column_pos.value = not options.preserve_column_pos.value end
}



function select_all_slots_in_sequencer()

  local song = renoise.song()
  --if 1st or last seq not selected then select all
  if song.sequencer.selection_range[1] ~= 1 or  
     song.sequencer.selection_range[2] ~= #song.sequencer.pattern_sequence then
    --set to whole range
    song.sequencer.selection_range = {1,#song.sequencer.pattern_sequence}
  else
    --set to current pattern
    song.sequencer.selection_range = {song.selected_pattern_index,song.selected_pattern_index}
  end
end



--bool- defined in following process functions
local single_track 


--Process slicers Current track or all tracks
-----------------------------------------------------------------------
--Current track
function process_single_track_in_pattern()
  single_track = true
  local slicer = ProcessSlicer(main)
  slicer:start()
end
-----------------------------------------------------------------------
--All tracks
function process_whole_track()
  single_track = false
  local slicer = ProcessSlicer(main)
  slicer:start()
end
-----------------------------------------------------------------------
--Whole Song
function process_whole_song()
  local slicer = ProcessSlicer(main_whole_song)
  slicer:start()
end

-----------------------------------------------------------------------
-----------------------------
--bubble sort
----------------------------- 
function modded_bubbleSort_left_align(t)
  
  --[i] is the column number
  --[i][1] is the note value for that column
  for i = 1, #t do
    for i2 = 1, #t do
      if t[i][1] < t[i2][1] then
        t[i],t[i2] = t[i2], t[i]
      end
    end
  end
end

-----------------------------
--Constants to filter in sort
local NOTE_OFF = 120
local EMPTY_NOTE = 121

-----------------------------
--bubble sort
----------------------------- 
function modded_bubbleSort(t)
  
  --[i] is the column number
  --[i][1] is the note value for that column
  for i = 1, #t do
    if t[i][1] < NOTE_OFF then --filter note_offs and empty_lines for i (only notes are compared and swapped)
      for i2 = 1, #t do
        if t[i2][1] < NOTE_OFF then --filter note_offs and empty_lines for i2 (only notes are compared and swapped)  
          if t[i][1] < t[i2][1] then
            t[i],t[i2] = t[i2], t[i]
          end
        end
      end
    end
  end
end
  
---------------------
--main
---------------------
function main()

  local song = renoise.song()
  
  local start_pattern
  local end_pattern
  
  if single_track then
    start_pattern = song.selected_pattern_index
    end_pattern = song.selected_pattern_index
  else
    start_pattern = 1
    end_pattern = #song.patterns
  end
 
  --song values  

  local pattern_iter = song.pattern_iterator
  local pattern_index = song.selected_pattern_index
  local track_index = song.selected_track_index
  --table and incrementors
  local pattern_notes = {}
  local selected_line = 1
  local current_line = 1
 
  --iterate
  for i = start_pattern,end_pattern do --only current pattern as too slow
   
    if (i % 3) == 0 then --only relevant for multiple patterns
      coroutine.yield()
    end 
    --reset counters for each pattern
    selected_line = 1
    current_line = 1
    --status
    if start_pattern == end_pattern then
      renoise.app():show_status("Re-Order Notes By Pitch Tool: (Re-Ordering `"..song.selected_track.name.."` Current Pattern: "..(i-1)..")")  
    else  
      renoise.app():show_status("Re-Order Notes By Pitch Tool: (Re-Ordering `"..song.selected_track.name.."` Pattern: "..(i-1).." / "..(end_pattern-1).."`)")   
    end
    --iterate track in pattern
    for pos,line in pattern_iter:note_columns_in_pattern_track(i,track_index) do
  
       pattern_notes[current_line] = { line.note_value, line.instrument_value, line.volume_value, line.panning_value, line.delay_value }
       current_line = current_line + 1
      --end of row
      if current_line == 13 then 
  
        --sort
        if not options.preserve_column_pos.value == true then
          modded_bubbleSort_left_align(pattern_notes)
        else
          modded_bubbleSort(pattern_notes)
        end
        --add in values
        local line = song.patterns[i].tracks[track_index].lines[selected_line]
      
        for i = 1,12 do
          line.note_columns[i]:clear()--clear current line
          line.note_columns[i].note_value = pattern_notes[i][1]--add sorted notes:
          line.note_columns[i].instrument_value = pattern_notes[i][2]
          line.note_columns[i].volume_value = pattern_notes[i][3]
          line.note_columns[i].panning_value = pattern_notes[i][4]
          line.note_columns[i].delay_value = pattern_notes[i][5] 
        end 
        --increment
        selected_line = selected_line + 1
        current_line = 1    
      end   
    end 
  end
renoise.app():show_status("Re-Order Notes By Pitch Tool: (Re-Order Complete)")
end


---------------------
--main whole song
---------------------
function main_whole_song()
  
  
  --song values  
  local song = renoise.song()
 
  --default range of current pattern
  local start_pattern = song.selected_pattern_index
  local end_pattern = song.selected_pattern_index
  
  --check if sequencer has selection then set range by it  
  if song.sequencer.selection_range[2] > 0 then
    start_pattern = song.sequencer.selection_range[1]
    end_pattern = song.sequencer.selection_range[2]
  end
    
  local seq_tracks = {}
  --get valid tracks i.e. miss groups/sends and master
  for i = 1, song.sequencer_track_count do
    if song.tracks[i].type == renoise.Track.TRACK_TYPE_SEQUENCER then
      table.insert(seq_tracks,i)
    end
  end
 
  --the pattern order gives the sometimes non-sequential order of pattern numbers in the sequencer  
  local pattern_order = {}
  for i = 1,#song.sequencer.pattern_sequence do
    table.insert(pattern_order, (song.sequencer.pattern_sequence[i]))
  end 
  
  local pattern_iter = song.pattern_iterator

  --table and incrementors
  local pattern_notes = {}
  local selected_line = 1
  local current_line = 1
 
  --loop
  for track = 1,#seq_tracks do
    
    for i = start_pattern,end_pattern do 
     
      if (i % 3) == 0 then --only relevant for multiple patterns
        coroutine.yield()
      end 
      --reset counters for each pattern
      selected_line = 1
      current_line = 1
      --status
      renoise.app():show_status("Re-Order Notes By Pitch Tool: (Re-Ordering Track: "..track.." / "..#seq_tracks..",  Sequencer Position: "..(i-1).." / "..(end_pattern-1).." Pattern no. "..(pattern_order[i]-1)..")") 
       
      --iterate track in pattern
      for pos,line in pattern_iter:note_columns_in_pattern_track(pattern_order[i],seq_tracks[track]) do
         --get each line in row values
         pattern_notes[current_line] = { line.note_value, line.instrument_value, line.volume_value, line.panning_value, line.delay_value }  
         current_line = current_line + 1
        
        --end of row
        if current_line == 13 then 
          --sort
           if not options.preserve_column_pos.value == true then
            modded_bubbleSort_left_align(pattern_notes)
          else
            modded_bubbleSort(pattern_notes)
          end
          --add in values
          local line = song.patterns[pattern_order[i]].tracks[seq_tracks[track]].lines[selected_line] 
        
          for i = 1,12 do
            line.note_columns[i]:clear()--clear current line
            line.note_columns[i].note_value = pattern_notes[i][1]--add sorted notes:
            line.note_columns[i].instrument_value = pattern_notes[i][2]
            line.note_columns[i].volume_value = pattern_notes[i][3]
            line.note_columns[i].panning_value = pattern_notes[i][4]
            line.note_columns[i].delay_value = pattern_notes[i][5] 
          end 
          --increment
          selected_line = selected_line + 1
          current_line = 1    
        end   
      end 
    end
  end
renoise.app():show_status("Re-Order Notes By Pitch Tool: (Re-Order Complete)")
end


